/* A wrapping of sample CUDA code to do 1D convolution in Mathematica. */

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include <mathlink.h>
#include <cuda.h>
#include <cufft.h>
#include <cutil.h>

// Complex data type
typedef float2 Complex; 
static __device__ __host__ inline Complex ComplexAdd(Complex, Complex);
static __device__ __host__ inline Complex ComplexScale(Complex, float);
static __device__ __host__ inline Complex ComplexMul(Complex, Complex);
static __global__ void ComplexPointwiseMulAndScale(Complex*, const Complex*, int, float);

// Padding functions
int PadData(const Complex*, Complex**, int,
            const Complex*, Complex**, int);

////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void cuConvolve1d(double *mm_kernel, long kernel_size,
                  double *mm_signal, long signal_size);

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[]) 
{
    int deviceCount = 0;
    CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceCount(&deviceCount));
    if (deviceCount == 0) {
        fprintf(stderr, "cutil error: no devices supporting CUDA.\n");
        exit(-1);
    }
    int dev = 0; // Default device 9600M GT on MacBook Pro
    cudaDeviceProp deviceProp;
    CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceProperties(&deviceProp, dev));
    if (deviceProp.major < 1) {
        fprintf(stderr, "cutil error: device does not support CUDA.\n");
        exit(EXIT_FAILURE);
    }
    CUDA_SAFE_CALL(cudaSetDevice(dev));      
    CUT_DEVICE_INIT(argc, argv);

    MLMain(argc, argv);

    fflush( stdout);
    fflush( stderr);
    exit(1);
}


:Begin:
:Function: cuConvolve1d
:Pattern:  CUDAConvolve1D[kernel:{__Real}, signal:{__Real}]
:Arguments: {kernel, signal}
:ArgumentTypes: {RealList, RealList}
:ReturnType: Manual
:End: 

void cuConvolve1d(double *mm_kernel, long kernel_size,
                  double *mm_signal, long signal_size)
{
    // Allocate host memory for the filter
    Complex* h_filter_kernel = (Complex*)malloc(sizeof(Complex) * kernel_size);
    // Initalize the memory for the filter
    for (unsigned int i = 0; i < kernel_size; ++i) {
        h_filter_kernel[i].x = (float)mm_kernel[i];
        h_filter_kernel[i].y = 0;
    }

    Complex* h_signal = (Complex*)malloc(sizeof(Complex) * signal_size);
    // Initalize the memory for the signal
    for (unsigned int i = 0; i < signal_size; ++i) {
        h_signal[i].x = (float)mm_signal[i];
        h_signal[i].y = 0;
    }

    // Pad signal and filter kernel
    Complex* h_padded_signal;
    Complex* h_padded_filter_kernel;
    int new_size = PadData(h_signal, &h_padded_signal, signal_size,
                           h_filter_kernel, &h_padded_filter_kernel, kernel_size);
    int mem_size = sizeof(Complex) * new_size;

    // Start timer
    unsigned int hTimer;
    CUT_SAFE_CALL( cutCreateTimer(&hTimer) );
    CUDA_SAFE_CALL( cudaThreadSynchronize() );
    CUT_SAFE_CALL( cutResetTimer(hTimer) );
    CUT_SAFE_CALL( cutStartTimer(hTimer) );
    
    // Allocate device memory for signal
    Complex* d_signal;
    CUDA_SAFE_CALL(cudaMalloc((void**)&d_signal, mem_size));

    // Copy host memory to device
    CUDA_SAFE_CALL(cudaMemcpy(d_signal, h_padded_signal, mem_size,
                              cudaMemcpyHostToDevice));

    // Allocate device memory for filter kernel
    Complex* d_filter_kernel;
    CUDA_SAFE_CALL(cudaMalloc((void**)&d_filter_kernel, mem_size));

    // Copy host memory to device
    CUDA_SAFE_CALL(cudaMemcpy(d_filter_kernel, h_padded_filter_kernel, mem_size,
                              cudaMemcpyHostToDevice));

    // CUFFT plan
    cufftHandle plan;
    CUFFT_SAFE_CALL(cufftPlan1d(&plan, new_size, CUFFT_C2C, 1));

    // Transform signal and kernel
    CUFFT_SAFE_CALL(cufftExecC2C(plan, (cufftComplex *)d_signal, (cufftComplex *)d_signal, CUFFT_FORWARD));
    CUFFT_SAFE_CALL(cufftExecC2C(plan, (cufftComplex *)d_filter_kernel, (cufftComplex *)d_filter_kernel, CUFFT_FORWARD));

    // Multiply the coefficients together and normalize the result
    ComplexPointwiseMulAndScale<<<32, 256>>>(d_signal, d_filter_kernel, new_size, 1.0f / new_size);

    // Check if kernel execution generated and error
    CUT_CHECK_ERROR("Kernel execution failed [ ComplexPointwiseMulAndScale ]");

    // Transform signal back
    CUFFT_SAFE_CALL(cufftExecC2C(plan, (cufftComplex *)d_signal, (cufftComplex *)d_signal, CUFFT_INVERSE));

    // Copy device memory to host
    Complex* h_convolved_signal = h_padded_signal;
    CUDA_SAFE_CALL(cudaMemcpy(h_convolved_signal, d_signal, mem_size,
                              cudaMemcpyDeviceToHost));
    
    // Stop timer
    CUDA_SAFE_CALL( cudaThreadSynchronize() );
    CUT_SAFE_CALL( cutStopTimer(hTimer) );
    double gpuTime = cutGetTimerValue(hTimer) /1000.;  // in seconds

    // Put complex list
    int dims[2]; 
    dims[0] = new_size;
    dims[1] = 2;

    MLPutFunction(stdlink, "List", 2);
    MLPutFunction(stdlink, "Apply", 3);
    MLPutSymbol(stdlink, "Complex");
    MLPutReal32Array(stdlink, (float*)h_convolved_signal, dims, (char **)0, 2);
    MLPutFunction(stdlink, "List", 1);
    MLPutInteger(stdlink, -2);
    
    // Put CUDA compute time
    MLPutReal64(stdlink, gpuTime);

    //Destroy CUFFT context
    CUFFT_SAFE_CALL(cufftDestroy(plan));

    // cleanup memory
    free(h_signal);
    free(h_filter_kernel);
    free(h_padded_signal);
    free(h_padded_filter_kernel);
    CUDA_SAFE_CALL(cudaFree(d_signal));
    CUDA_SAFE_CALL(cudaFree(d_filter_kernel));
}


// Pad data
int PadData(const Complex* signal, Complex** padded_signal, int signal_size,
            const Complex* filter_kernel, Complex** padded_filter_kernel, int filter_kernel_size)
{
    int minRadius = filter_kernel_size / 2;
    int maxRadius = filter_kernel_size - minRadius;
    int new_size = signal_size + maxRadius;
    
    // Pad signal
    Complex* new_data = (Complex*)malloc(sizeof(Complex) * new_size);
    memcpy(new_data +           0, signal,              signal_size * sizeof(Complex));
    memset(new_data + signal_size,      0, (new_size - signal_size) * sizeof(Complex));
    *padded_signal = new_data;
    
    // Pad filter
    new_data = (Complex*)malloc(sizeof(Complex) * new_size);  
    memcpy(new_data +                    0, filter_kernel + minRadius,                       maxRadius * sizeof(Complex));
    memset(new_data +            maxRadius,                         0, (new_size - filter_kernel_size) * sizeof(Complex));   
    memcpy(new_data + new_size - minRadius,             filter_kernel,                       minRadius * sizeof(Complex));
    *padded_filter_kernel = new_data;
    
    return new_size;
}


////////////////////////////////////////////////////////////////////////////////
// Complex operations
////////////////////////////////////////////////////////////////////////////////

// Complex addition
static __device__ __host__ inline Complex ComplexAdd(Complex a, Complex b)
{
    Complex c;
    c.x = a.x + b.x;
    c.y = a.y + b.y;
    return c;
}

// Complex scale
static __device__ __host__ inline Complex ComplexScale(Complex a, float s)
{
    Complex c;
    c.x = s * a.x;
    c.y = s * a.y;
    return c;
}

// Complex multiplication
static __device__ __host__ inline Complex ComplexMul(Complex a, Complex b)
{
    Complex c;
    c.x = a.x * b.x - a.y * b.y;
    c.y = a.x * b.y + a.y * b.x;
    return c;
}

// Complex pointwise multiplication
static __global__ void ComplexPointwiseMulAndScale(Complex* a, const Complex* b, int size, float scale)
{
    const int numThreads = blockDim.x * gridDim.x;
    const int threadID = blockIdx.x * blockDim.x + threadIdx.x;
    for (int i = threadID; i < size; i += numThreads)
        a[i] = ComplexScale(ComplexMul(a[i], b[i]), scale);     
} 
